/*
 * ---------------------------------------------------------------------------
 * This file defines constants, functions, ... used to calculate the part of
 * O2, He, N2 gas which must be filled into the tank to get the wanted gasmix.
 * author : Mstl Georg
 * date   : 22.11.2006
 * ---------------------------------------------------------------------------
 */
 

#include "gasmixer_defs.h"

#include <stdint.h>
#include <stdio.h>
#include <stdbool.h>
#include <avr/eeprom.h>

 
/* 
 * -------------------------------------------------------------------------
 * internal function declaration
 * -------------------------------------------------------------------------
 */

inline float changeVToPCALC(uint8_t p, uint8_t V);

inline uint8_t checkPDiffCALC(float* diff);

inline uint8_t detNewSrcTankPCALC(float pp, uint8_t V);

fill_instr_t calcGasAmountCALC(float* pp_diff, uint8_t p_diff);


/* 
 * -------------------------------------------------------------------------
 * internal function declaration
 * -------------------------------------------------------------------------
 */

inline float changeVToPCALC(uint8_t p, uint8_t V) {

	if(V == 0 || p == 0) return 0.0;
		
	return p * (V/100.0);

}//changeVToPCALC


inline uint8_t checkPDiffCALC(float* diff) {

	uint8_t i;
	
	for(i=0; i<3 && *diff++ >= 0.0; i++);
	
	return i;
	
}//checkPDiffCALC


inline uint8_t detNewSrcTankPCALC(float pp, uint8_t V) {

  //p has to be rounded down every time so cast (uint8_t) is possible
	return (uint8_t) (pp*100/V);

}//detNewSrcTankPCALC


fill_instr_t calcGasAmountCALC(float* pp_diff, uint8_t p_diff) {

	extern cfg_par_t cfg_par;
	
	cfg_par_t par;
	
	float w1, w2, w3; //weights of topoff for O2, He, N2
	float p1, p2, p3; //p difference needed for O2, He, N2
	float p;          //p differnce between dest and src tank
	float x, y;       //p needed to fill in for O2, He
	
	fill_instr_t instr = { {0, 0, 0}, 0};
	
	//get TopOff
	eeprom_read_block(&par, &cfg_par, sizeof(cfg_par_t));
	
	//determine parameters/constants
	w1 = par.topoff.per_O2/100.0;
	w2 = par.topoff.per_He/100.0;
	w3 = par.topoff.per_N2/100.0;
	p1 = *(pp_diff + IDX_O2);
	p2 = *(pp_diff + IDX_HE);
	p3 = *(pp_diff + IDX_N2);
	p  = p_diff;
	
	//determine amount of O2 [bar]
	//x = (1 / (1 - w1 - (w1*w2/(1-w2)))) * 
	//    (p1 - p*w1 + 1/(1-w2)*(p2*w1 - p*w1*w2))
	x = ((1 / (1 - w1 - (w1*w2/(1-w2)))) * 
	    (p1 - p*w1 + 1/(1-w2)*(p2*w1 - p*w1*w2)));
	
	//determine amount of He [bar]
	//y = (1 / (1-w2)) * (p2 + (x - p) * w2)
	y = (1 / (1-w2)) * (p2 + (x - p) * w2);

	//round calculated values to integer value
  instr.amount.am_O2  = (uint8_t) (x + 0.5);
	instr.amount.am_He  = (uint8_t) (y + 0.5);
	instr.amount.am_Top = p_diff - instr.amount.am_O2 - instr.amount.am_He;
	
  return instr;
	
}//calcGasAmountCALC

 
/* 
 * -------------------------------------------------------------------------
 * general function definition
 * -------------------------------------------------------------------------
 */
 
fill_instr_t calcGasmixCALC(tank_t src_tank, tank_t dest_tank) {
	
	uint8_t i;
	
	const uint8_t src_V[]  = {src_tank.gas.per_O2, src_tank.gas.per_He, 
	                          src_tank.gas.per_N2};
	const uint8_t dest_V[] = {dest_tank.gas.per_O2, dest_tank.gas.per_He, 
	                          dest_tank.gas.per_N2};
	
	float pp_src[]  = {0.0, 0.0, 0.0}; //partial pressure src tank (O2, He, N2)
	float pp_dest[] = {0.0, 0.0, 0.0}; //- " -            dest tank (O2, He, N2)
	float pp_diff[] = {0.0, 0.0, 0.0}; //partial pressure difference 
	                                   //(dest - src)
	
	fill_instr_t instr = { {0, 0, 0}, 0};
	
	//change volume into partial pressure -> because of ideal gases the [%] part
	//of the volume of a specific gas in the mix is directly related to the 
	//partial pressure
	for(i=0; i<3; i++) {
		pp_src[i]  = changeVToPCALC(src_tank.p, src_V[i]);
		pp_dest[i] = changeVToPCALC(dest_tank.p, dest_V[i]);
  }

	//calculate pressure difference of dest and src tank for each single gas 
	//(O2, He, N2)
	//-> if difference is negative than src tank must be emptied until gas mix 
	//is possible
	do {
	
		for(i=0; i<3; i++) 
			pp_diff[i] = pp_dest[i] - pp_src[i];
		
		if((i=checkPDiffCALC(&pp_diff[0])) < 3) {
		
			src_tank.p = detNewSrcTankPCALC(pp_dest[i], src_V[i]);
			for(i=0; i<3; i++) 
				pp_src[i] = changeVToPCALC(src_tank.p, src_V[i]);

		}//if
		
	} while(checkPDiffCALC(&pp_diff[0]) < 3);	
 
	//gasmix is possible calculate single p (O2, He, TopOff)
	//-> checking of storing tank p diff in calling subroutine
	instr = calcGasAmountCALC(&pp_diff[0], dest_tank.p - src_tank.p);
	instr.p = src_tank.p;
	
	return instr;
 
}//calcGasmixCALC

